#!/usr/bin/env python3

from emuserema import Emuserema
from sys import argv
from IPython import embed


def main():
    emuserema = Emuserema()
    if len(argv) > 1:
        if argv[1] == '-t':
            print("Configuration parsing was successful.")
            exit(0)
    embed()
    emuserema.render()


if __name__ == '__main__':
    main()



from emuserema.services import SSHservice
from emuserema.plugin_manager import Plugin
from emuserema.utils import makedir_getfd


class PuttyRenderer(Plugin):
    def config(self):
        self.description = 'PuTTY registry generator'
        self.registries = {}

    def render_putty(self):
        for key, service in self.services.items():
            if isinstance(service, SSHservice):
                putty_config = self.open_registry(service.path[0])
                print("[HKEY_CURRENT_USER\\Software\\SimonTatham\\PuTTY\\Sessions\\%s]" %
                    service.tag, file=putty_config)
                for k, v in self.render_kv(service):
                    print("\"%s\"=%s" % (k, v), file=putty_config)
                print('', file=putty_config)
        self.close_registries()

    def render_kv(self, service):
        rv = []
        for key, value in service.kwargs.items():
            if key[0] == '_':  # Blindly copy all the non meta settings
                continue
            key = key.lower()
            if key == 'user':
                rv.append(('UserName', '"%s"' % value))
            elif key == 'port':
                rv.append(('PortNumber', 'dword:%08x' % value))
        port_forwardings = []
        if service.forwards['local']:
            #print(service.forwards['local'])
            port_forwardings += map(lambda x: "L" + x[0] + "=" + x[1][0], service.forwards['local'].items())
        if service.forwards['remote']:
            port_forwardings += map(lambda x: "R" + x[0] + "=" + x[1][0], service.forwards['remote'].items())
        if service.forwards['dynamic']:
            port_forwardings += map(lambda x: "D" + x, service.forwards['dynamic'])
        rv.append(('PortForwardings', '"' + ','.join(port_forwardings) + '"'))
        for k, v in service.putty_settings:
            rv.append((k, v))
        return rv

    def open_registry(self, world_name):
        if world_name not in self.registries:
            self.registries[world_name] = makedir_getfd("%s/%s.reg" % (self._config['output_dir'], world_name))
            print("Windows Registry Editor Version 5.00\n", file=self.registries[world_name])
            print("[HKEY_CURRENT_USER\\Software\\SimonTatham\\PuTTY\\Sessions]\n", file=self.registries[world_name])

        return self.registries[world_name]

    def close_registries(self):
        for key, registry in self.registries.items():
            registry.close()

    def run(self, **kwargs):
        self.services = kwargs['services']
        self.worlds = kwargs['worlds']
        self.render_putty()



#import ruamel.yaml

#from ruamel.std.pathlib import Path
#from ruamel.yaml import YAML, version_info, dump, YAMLError, SafeLoader, RoundTripLoader
from ruamel.yaml import YAML, YAMLError, dump, SafeLoader


class EmuseremaRelativeSafeLoader(SafeLoader):
    def __init__(self, definitions_directory=None):
        self._definitions_directory = definitions_directory

    def construct_python_tuple(self, cons, node):
        return tuple(cons.construct_sequence(node))

    # adapted from http://code.activestate.com/recipes/577613-yaml-include-support/
    def yaml_include(self, cons, node):
        y = cons.loader
        yaml = YAML(typ=y.typ, pure=y.pure)  # same values as including YAML
        yaml.composer.anchors = cons.composer.anchors
        with open(self._definitions_directory + '/' + node.value, 'r') as infile:
            return yaml.load(infile)


class EmuseremaYamlLoader(object):
    def __init__(self, definitions_directory=None):
        self._definitions_directory = definitions_directory

        self.yaml = YAML(typ='safe', pure=True)
        self.yaml.default_flow_style = False

        def my_compose_document(self):
            self.parser.get_event()
            node = self.compose_node(None, None)
            self.parser.get_event()
            # self.anchors = {}    # <<<< commented out
            return node

        self.yaml.Composer.compose_document = my_compose_document

        loader = EmuseremaRelativeSafeLoader(self._definitions_directory)

        self.yaml.Constructor.add_constructor('!include', loader.yaml_include)
        self.yaml.Constructor.add_constructor('tag:yaml.org,2002:python/tuple', loader.construct_python_tuple)

    def loadyaml(self, filename):
        yamldata = None
        with open(self._definitions_directory + '/' + filename, 'r') as stream:
            try:
                yamldata = self.yaml.load(stream)
        #        print(dump(t, default_flow_style=False))
        #        print type(t)
            except YAMLError as exc:
                print(exc)

        if not yamldata:
            print(yamldata)
            raise ValueError("Failed to parse YAML Data")
        return yamldata

    def dumpyaml(self, data, filename):

        with open(self._definitions_directory + '/' + filename, 'w') as stream:
            dump(data, stream, default_flow_style=False)




#!/usr/bin/env python3

from os.path import expanduser
from sys import exit, argv
from emuserema.plugin_manager import PluginManager
from emuserema.services import AbstractService
from emuserema.world import World
from emuserema.yamlloader import EmuseremaYamlLoader
from emuserema.utils import traverse
from emuserema.services import *
from emuserema.redirects import RedirectFactory
# if yaml dumping is to be enabled:
#from yamlloader import dump
#from sys import stdout

#CFGDIR = PREFIX + '/etc'
#LIBDIR = PREFIX + '/local/lib/emuserema'
#DESTINATION_HOSTS_FILE = PREFIX + '/etc/emuserema/per-destination-commands'
#from IPython import embed


class Emuserema(object):
    def __init__(self, definitions_directory='~/.config/emuserema'):
        self._definitions_directory = expanduser(definitions_directory)

        self.yamlloader = EmuseremaYamlLoader(self._definitions_directory)
        self.redirect_factory = RedirectFactory(definitions_directory=self._definitions_directory,
                yamlloader=self.yamlloader)
        self.service_factory = ServiceFactory()
        self.services = {}
        self.worlds = {}

        self.load()

    def __del__(self):
        self.commit()

    def service_builder(self, path, obj):
        if path and path[0] == 'defaults':
            return obj
        if isinstance(obj, dict):
            if '_type' in obj:
                if self.service_factory.supported(obj):
                    service = self.service_factory.create(path, **obj)
                    self.services[service.tag] = service
                    return service
                else:
                    raise NotImplementedError("Object type of '%s' is not supported" % obj['_type'])
        return obj

    def via_resolver(self, path, service):
        if path and path[0] == 'defaults':
            return service
        if isinstance(service, AbstractService):
            if service.via is None:
                return service
            service.via = self.services[service.via]
        return service

    def create_worlds(self, path, service):
        if path and path[0] == 'defaults':
            return service
        if isinstance(service, AbstractService):
            if service.world not in self.worlds:
                self.worlds[service.world] = World(service.world)
            #service.world=worlds[service.world]
            self.worlds[service.world].append(service)

        return service

    def redirector(self, path, service):
        if path and path[0] == 'defaults':
            return service
        if isinstance(service, AbstractService):
            if service.via is None:
                return service
            service.process_proxy(service.via, self.redirect_factory)

        return service

    def load(self):
        self.data = self.yamlloader.loadyaml('emuserema.yaml')

        self.data = traverse(self.data, callback=self.service_builder)
        self.data = traverse(self.data, callback=self.via_resolver)
        self.data = traverse(self.data, callback=self.redirector)
        self.data = traverse(self.data, callback=self.create_worlds)
        #dump(self.data, stdout, default_flow_style=False)

    def render(self):
        config = self.yamlloader.loadyaml('config.yaml')

        renderer_plugins = PluginManager('emuserema.plugins.renderers', config=config['plugins']['renderers'])

        renderer_plugins.run_selected(filter(lambda name: config['plugins']['renderers'][name]['enabled'] is True,
            config['plugins']['renderers']), worlds=self.worlds, services=self.services)

    def populate_ansible_inventory(self, inventory=None, world='default'):
        inventory.add_host('teszt')

        world = self.worlds[world]
        for service in self.services:
            service = self.services[service]
            if self.worlds[service.world] == world:
                if isinstance(service, SSHservice) or isinstance(service, DummyService) or '_ansible_hostvars' in service.kwargs:
                    inventory.add_host(service.tag)
                    inventory.set_variable(service.tag, 'ansible_ssh_common_args',
                        '-F /home/e/.ssh/configs/%s' % service.world)
                    #we must fix this somehow
                    #inventory.set_variable(service.tag, 'ansible_ssh_executable',
                    #    'SSH_AUTH_SOCK=/home/e/.ssh/agents/%s /usr/bin/ssh' % service.world)
                    if '_ansible_hostvars' in service.kwargs:
                        for key, value in service.kwargs['_ansible_hostvars'].items():
                            inventory.set_variable(service.tag, key, value)
                    if '_ansible_groups' in service.kwargs:
                        for group in service.kwargs['_ansible_groups']:
                            group = group.lower().replace(' ', '_')
                            group = inventory.add_group(group)
                            inventory.add_child(group, service.tag)
                    for group in service.path[1:-1] + ['emuserema_' + '__'.join(service.path[1:-1])]:
                        group = group.lower().replace(' ', '_')
                        group = inventory.add_group(group)
                        inventory.add_child(group, service.tag)
#        embed()
        pass

    def commit(self):
        self.redirect_factory.commit()


def main():
    emuserema = Emuserema()
    if len(argv) > 1:
        if argv[1] == '-t':
            print("Configuration parsing was successful.")
            exit(0)
    emuserema.render()


if __name__ == '__main__':
    main()




from emuserema.services import SSHservice
from emuserema.plugin_manager import Plugin
from emuserema.utils import makedir_getfd, cleanup_files


class OpenSSHRenderer(Plugin):
    def config(self):
        self.description = 'openssh client configuration files renderer'
        self.ssh_config_files = {}

    def cleanup(self):
        cleanup_files(self._config['output_dir'])

    def render_openssh(self):
        for key, service in self.services.items():
            #service=self.services[key]
            if isinstance(service, SSHservice):
                ssh_config_file = self.open_ssh_config_file(service.world)
                print("\n###\n### %s\n###\n" % service.tag, file=ssh_config_file)
                print("Host %s" % service.tag, file=ssh_config_file)
                print("\n".join(service.pb), file=ssh_config_file)
                #for k in self.render_kv(service):
                #    if type(k) == tuple:
                #        print("\t%s = %s" % (k[0], k[1]), file=ssh_config_file)
                #    elif type(k) == str:
                #        print("\t"+k, file=ssh_config_file)

        self.close_ssh_config_files()

    def render_kv(self, service):
        rv = []
        for key, value in service.kwargs.items():
            if key[0] != '_':  # Blindly copy all the non meta settings
                rv.append((key, value))
        for bind_hostport, target_hostport in service.forwards['local'].items():
            if target_hostport[1]:
                rv.append('#' + target_hostport[1])
            rv.append(('LocalForward', "%s %s" % (bind_hostport, target_hostport[0])))
        for bind_hostport, target_hostport in service.forwards['remote']:
            if target_hostport[1]:
                rv.append('#' + target_hostport[1])
                rv.append(('RemoteForward', "%s %s" % (bind_hostport, target_hostport[0])))
        for bind_hostport in service.forwards['dynamic']:
            if target_hostport[1]:
                rv.append('#' + target_hostport[1])
                rv.append(('DynamicForward', "%s" % (bind_hostport, target_hostport[0])))
        return rv

    def open_ssh_config_file(self, world_name):
        if world_name not in self.ssh_config_files:
            self.ssh_config_files[world_name] = makedir_getfd("%s/%s" % (self._config['output_dir'], world_name))
            try:
                with open('templates/ssh_config_prefix', 'r') as preamble:
                    for line in preamble:
                        self.ssh_config_files[world_name].write(line)
            except FileNotFoundError:
                pass

        return self.ssh_config_files[world_name]

    def close_ssh_config_files(self):
        for key, ssh_config_file in self.ssh_config_files.items():
            try:
                with open('templates/ssh_config_postfix', 'r') as postfix:
                    for line in postfix:
                        ssh_config_file.write(line)
            except FileNotFoundError:
                pass
            ssh_config_file.close()

    def run(self, **kwargs):
        self.services = kwargs['services']
        self.worlds = kwargs['worlds']
        self.render_openssh()



from emuserema.services import URLservice
from emuserema.plugin_manager import Plugin
from emuserema.utils import service_paths_to_tree, makedir_getfd
import json


class JsTreeRenderer(Plugin):
    def config(self):
        self.description = 'jstree URL launcher site renderer'

    def render_jstree(self, world):
        #tree=[]
        tree = []
        lst = []

        for service in world.services.values():
            if isinstance(service, URLservice):
                lst.append((service.path[1:], service))
        tree = service_paths_to_tree(lst)
        if tree:

            with makedir_getfd("%s/%s.html" % (self._config['output_dir'], world.name)) as dst:
                try:
                    with open('templates/jstree-html-prefix.html', 'r') as preamble:
                        for line in preamble:
                            dst.write(line)
                except FileNotFoundError:
                    pass

                json.dump(obj=tree, fp=dst, sort_keys=True, indent=2)

                try:
                    with open('templates/jstree-html-postfix.html', 'r') as preamble:
                        for line in preamble:
                            dst.write(line)
                except FileNotFoundError:
                    pass

    def run(self, **kwargs):
        """The actual implementation of the identity plugin is to just return the
        argument
        """
        self.services = kwargs['services']
        self.worlds = kwargs['worlds']
        for world in self.worlds:
            self.render_jstree(self.worlds[world])




USE_PANGO_TAGS=1
STAGE3_PATH='~/bin/emuserema-ssh-wrapper.sh'
from abc import ABC, abstractmethod

#FOR URLservice support:
from urllib.parse import urlparse


class AbstractService(ABC):
    def __init__(self, path=[], **kwargs):
        self.pb = []
        self.tag = path[-1]
        self.world = path[0]
        self.path = path
        self.kwargs = kwargs
        #self.label="%s\t" % tag
        self.label = ""
        self.via = None

        self.config()

    @abstractmethod
    def config(self):
        raise NotImplementedError

    @abstractmethod
    def process_proxy(self, via_service, redirect_factory):
        raise NotImplementedError

    def __repr__(self):
        return("<{}({!r})>".format(self.__class__.__name__, self.tag))


class DummyService(AbstractService):
    def __init__(self, path=[], **kwargs):
        self.pb = []
        self.tag = path[-1]
        self.world = path[0]
        self.path = path
        self.kwargs = kwargs
        #self.label="%s\t" % tag
        self.label = ""
        self.via = None

        self.config()

    def config(self):
        pass

    def process_proxy(self, via_service, redirect_factory):
        pass

    def __repr__(self):
        return("<{}({!r})>".format(self.__class__.__name__, self.tag))


class SSHservice(AbstractService):

    #This is to support redirection of hosts that are SCB monitored and behind restricted SSH terminal server gateways
    @property
    def world(self):
        if self.via and not isinstance(self.via, SSHoverSCBservice):
            return self.via.world
        return self._world

    @property
    def realworld(self):
        return self._world

    @world.setter
    def world(self, world):
        self._world=world

    def config(self):
        self.putty_settings=[]
        self.forwards={
            'local': {},
            'remote': {},
            'dynamic': {}
        }

        if 'hostname' not in map(lambda x: x.lower(), self.kwargs.keys()):
            self.kwargs['HostName'] = self.tag
        for key, value in self.kwargs.items():
            if key[0]!='_': #Blindly copy all the non meta settings
                self.pb.append("\t%s = %s" % (key, self.kwargs[key]))
            else:
                key=key.lower()
                if key == '_via':
                    self.via = value
                elif key == '_forward':
                    #print self.kwargs[key]
                    if 'local' in self.kwargs[key]:
                        for bind_hostport, target_hostport in self.kwargs[key]['local'].items():
                            self.add_local_forward(bind_hostport, target_hostport, comment="added from YAML")
                            self.pb.append('\tLocalForward = %s %s' % (bind_hostport, target_hostport))
                    if 'remote' in self.kwargs[key]:
                        for bind_hostport, target_hostport in self.kwargs[key]['remote'].items():
                            self.add_remote_forward(bind_hostport, target_hostport, comment="added from YAML")
                            self.pb.append('\tRemoteForward = %s %s' % (bind_hostport, target_hostport))
                    if 'dynamic' in self.kwargs[key]:
                        for bind_hostport in self.kwargs[key]['dynamic']:
                            self.add_dynamic_forward(bind_hostport, target_hostport, comment="added from YAML")
                            self.pb.append('\tDynamicForward = %s %s' % (bind_hostport))
                elif key == '_hostkeyalias':
                    self.pb.append('\tHostKeyAlias = %s' % (self.tag))
        self.add_label()

    def add_label(self):
        if '_terminal' not in self.kwargs:
            #self.label="%s" % self.tag
            if USE_PANGO_TAGS:
                self.label+=" <span font_size='x-small' fgcolor='#a0a080'>as %s</span>" % (self.kwargs['User'])
            self.label+="\turxvtc -tn rxvt-unicode -T \"SSH: %s@%s\" -insecure -e %s %s %s" % (self.kwargs['User'], self.tag, STAGE3_PATH, self.tag, self.tag)
        pass

    def process_proxy(self, via_service, redirect_factory):
        if type(via_service) == SSHservice:
            self.pb.append('\tProxyCommand = ssh -oLogLevel=QUIET -F ~/.ssh/configs/{0} -W \'[%h]\':%p {1}'.format(via_service.world, via_service.tag))
            ##print(self.kwargs.keys())
            self.putty_settings.append(("ProxyTelnetCommand",'"plink %s -nc %s:%s\\n"' % (via_service.tag, self.kwargs['HostName'], self.kwargs['Port'])))
            #if type(via_service) == SSHoverSCBservice:
            #    self.world = via_service.tag
        elif type(via_service) == SSHoverSCBservice:
            self.world = via_service.tag
            self.label = ''
            if USE_PANGO_TAGS:
                self.label+=" <span font_size='x-small' fgcolor='#a0a080'>as %s</span>" % (self.kwargs['User'])
                self.label+=" <span font_size='xx-small' font_style='italic' fgcolor='#a0a080'>(via %s)</span>" % (via_service.tag)
            self.label+="\turxvtc -tn rxvt-unicode -T \"SSH: %s@%s via %s\" -insecure -xrm \"urxvt.answerbackString: clear;exec /usr/bin/ssh -F ~/.ssh/configs/%s %s\\\\n\" -e %s %s %s" % (self.kwargs['User'], self.tag, via_service.tag, via_service.tag, self.tag, STAGE3_PATH, via_service.tag, self.tag)
            self.putty_settings.append(("HostName",'"%s"'% via_service.tag))
            self.putty_settings.append(("Answerback",'"clear;exec /usr/bin/ssh -F ~/.ssh/configs/%s %s^M"' % (via_service.tag, self.tag)))
            #self.putty_settings.append(("ProxyTelnetCommand",'"plink %s -nc %s:%s\\n"' % (via_service.tag, self.kwargs['HostName'], self.kwargs['Port'])))

    def add_local_forward(self, bind_hostport, target_hostport, comment=None):
        self.forwards['local'][bind_hostport]=(target_hostport, comment)

    def add_remote_forward(self, bind_host, bind_port, target_host, target_port, comment=None):
        self.forwards['remote'][bind_hostport]=(target_hostport, comment)

    def add_dynamic_forward(self, bind_host, bind_port, target_host, target_port, comment=None):
        self.forwards['dynamic'][bind_hostport]=comment


class SSHoverSCBservice(SSHservice):
    pass


class URLservice(AbstractService):
    def config(self):
        if '_via' in self.kwargs:
            self.via = self.kwargs['_via']
        self.url = self.kwargs['url']
        self.original_url = self.kwargs['url']

    def process_proxy(self, via_service, redirect_factory):
        purl = urlparse(self.url)
        (bindip, bindport) = redirect_factory.allocate(via_service, self)
        dport = purl.port

        if dport is None:
            if purl.scheme == 'http':
                dport = 80
            elif purl.scheme == 'https':
                dport = 443
            else:
                raise NotImplementedError("URL scheme '%s' is not supported" % purl.scheme)

        via_service.pb.append('\t# EMUSEREMA added implicit redirection for service "%s" at URL: %s' % (self.tag, self.url) )
        via_service.pb.append('\tLocalForward = %s:%s %s' % (bindip, bindport, "%s:%s" % (purl.hostname, dport)))
        via_service.add_local_forward("%s:%s" % (bindip, bindport), "%s:%s" % (purl.hostname, dport), '# EMUSEREMA added implicit redirection for service "%s" at URL: %s' % (self.tag, self.url))

        self.original_url = self.url
        self.url = purl._replace(netloc = '%s:%s' % (bindip, bindport)).geturl()


class RDPservice(AbstractService):
    def config(self):
        if '_via' in self.kwargs:
            self.via = self.kwargs['_via']

    def process_proxy(self, via_service, redirect_factory):
        self._original_full_address=self.kwargs["full address"]
        (bindip, bindport) = redirect_factory.allocate(via_service, self)
        via_service.pb.append('\t# EMUSEREMA added implicit redirection for RDP service "%s" at listening address: %s' % (self.tag, self._original_full_address))
        via_service.pb.append('\tLocalForward = %s:%s %s' % (bindip, bindport, "%s" % (self._original_full_address[2:])))
        self.kwargs["full address"] = "s:%s:%s" % (bindip, bindport)


class VNCservice(AbstractService):
    def config(self):
        if '_via' in self.kwargs:
            self.via = self.kwargs['_via']

    def process_proxy(self, via_service, redirect_factory):
        bindip, bindport = redirect_factory.allocate(via_service, self)

        try:  # Direct port definition
            self.kwargs['Host'].index('::')
            targetip, targetport = self.kwargs['Host'].split('::')
        except ValueError:  # VNC display offset definition
            try:
                self.kwargs['Host'].index(':')
                targetip, targetport = self.kwargs['Host'].split(':')
                targetport = int(targetport) + 5900
            except ValueError:  # Fallback to port 5900 when no port or display number specified
                targetip = self.kwargs['Host']
                targetport = 5900

        via_service.pb.append('\t# EMUSEREMA added implicit redirection for VNC service "%s" at listening address: %s' % (self.tag, self.kwargs['Host']) )
        self.original_host = self.kwargs['Host']
        via_service.pb.append('\tLocalForward = %s %s:%s' % (self.kwargs['Host'], targetip, targetport))
        self.kwargs['Host'] = '%s:%s' % (bindip, bindport)


class ServiceFactory(object):
    def __init__(self):
        self.builders={
            'ssh': SSHservice,
            'ssh_over_scb': SSHoverSCBservice,
            'url': URLservice,
            'dummy': DummyService,
            'rdp': RDPservice,
            'vnc': VNCservice
        }

    def create(self, path, **kwargs):
        return self.builders[kwargs['_type']](path, **kwargs)

    def supported(self, obj):
        return obj['_type'] in self.builders.keys()




from emuserema.plugin_manager import Plugin
from emuserema.services import VNCservice
from emuserema.utils import makedir_getfd

# for RealVNC session file UUID creation
from uuid import UUID, uuid5
vnc_namespace = UUID('d1982572-cbf4-5bdc-b52e-a03c843e2f4d')


class RealVNCRenderer(Plugin):
    def config(self):
        self.description = 'RealVNC profile files renderer'

    def render_realvnc(self):
        for key, service in self.services.items():
            if isinstance(service, VNCservice):
                uuid = uuid5(vnc_namespace, service.tag)
                service.kwargs['Uuid'] = uuid
                if 'FriendlyName' not in service.kwargs:
                    service.kwargs['FriendlyName'] = service.tag
                with makedir_getfd('%s/%s' % (self._config['output_dir'], str(uuid))) as dst:
                    for key, value in service.kwargs.items():
                        if key[0] != '_':
                            print("%s=%s" % (key, value or ''), file=dst)

    def run(self, **kwargs):
        """The actual implementation of the identity plugin is to just return the
        argument
        """
        self.services = kwargs['services']
        self.worlds = kwargs['worlds']
        self.render_realvnc()




from emuserema.services import RDPservice
from emuserema.plugin_manager import Plugin
from emuserema.utils import makedir_getfd, cleanup_dir


class RDPTreeRenderer(Plugin):
    def config(self):
        self.description = 'mstsc.exe ".rdp" profile files renderer'

    def cleanup(self):
        cleanup_dir(self._config['output_dir'])

    def render_rdp(self):
        for key, service in self.services.items():
            if isinstance(service, RDPservice):
                with makedir_getfd("%s/%s/%s" % (self._config['output_dir'], '/'.join(service.path))) as dst:
                    for key, value in service.kwargs.items():
                        if key[0] != '_':
                            print("%s:%s" % (key, value), file=dst)

    def run(self, **kwargs):
        self.services = kwargs['services']
        self.worlds = kwargs['worlds']
        self.render_rdp()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Endre Szabo
# GNU General Public License v2.0 (see COPYING or https://www.gnu.org/licenses/gpl-2.0.txt)

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
    name: emuserema
    plugin_type: inventory
    short_description: EMUSEREMA inventory source
    description:
        - Builds inventory from specific EMUSEREMA worlds
    extends_documentation_fragment:
      - constructed
    version_added: 2.9.2
    author: Endre Szabo (@endreszabo)
    options:
        world:
            description: EMUSEREMA world name to render
            required: true
    requierements: a working EMUSEREMA environment
    seealso:
      - name: EMUSEREMA project homepage
        description: EMUSEREMA project homepage with extensive documentations and examples.
        link: https://end.re/emuserema/
'''

EXAMPLES = r'''
---
plugin: emuserema
world: home
'''

#from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.plugins.inventory import BaseInventoryPlugin, Constructable
from sys import path

path.append('/home/e/git')

from emuserema import Emuserema

class InventoryModule(BaseInventoryPlugin, Constructable):
    ''' Host inventory parser for ansible using native EMUSEREMA. '''

    NAME = 'emuserema'

    def __init__(self):
        super(InventoryModule, self).__init__()

    def verify_file(self, path):
        ''' return true/false if this is possibly a valid file for this plugin to consume '''
        valid = False
        if super(InventoryModule, self).verify_file(path):
            # base class verifies that file exists and is readable by current user
            if path.endswith('emuserema.yaml'):
                valid = True
        return valid

    def parse(self, inventory, loader, path, cache=True):

        super(InventoryModule, self).parse(inventory, loader, path)

        config_data = self._read_config_data(path)

        # set _options from config data
        self._consume_options(config_data)

        self.emuserema = Emuserema()

        return self.emuserema.populate_ansible_inventory(inventory=inventory, world=self.get_option('world'))



from emuserema.services import *

class World(object):
    """ Obsolete class, only one renderer uses it """
    def __init__(self, name):
        self.name = name
        self.redirects=[]
        self.services={}
        self.tree={}
        self.service_index={}

    def append(self, service):
        self.services[service.tag] = service

    def render(self):
        pass

    def __repr__(self):
        return "<World '%s'>" % self.name

def __str__(self):
    return self.name

def append_redirect(self, via_service, target_service):
    self.redirects.append(self.redirect_factory.allocate(via_service, target_service))
    pass



# This section is mostly based on works of Guido Diepen as seen at
# https://www.guidodiepen.nl/2019/02/implementing-a-simple-plugin-framework-in-python/

import pkgutil
import os
import inspect


class Plugin(object):
    def __init__(self, config={}):
        self.description = 'UNKNOWN'
        self._config = config
        self.config()
        if 'cleanup' in self._config and self._config['cleanup'] is True:
            self.cleanup()

    def cleanup(self):
        raise NotImplementedError

    def config(self):
        raise NotImplementedError

    def render(self, kwargs):
        raise NotImplementedError


class PluginManager(object):
    def __init__(self, plugin_package, config):
        self.plugin_package = plugin_package
        self.config = config
        self.reload_plugins()

    def reload_plugins(self):
        self.plugins = []
        self.seen_paths = []
        self.walk_package(self.plugin_package)

    def run_all(self, kwargs):
        rv = []
        for plugin in self.plugins:
            rv.append(plugin.run(**kwargs))
        return rv

    def run_selected(self, enabled_plugins, **kwargs):
        rv = []
        enabled_plugins = list(enabled_plugins)
        for plugin in self.plugins:
            if plugin.__class__.__module__.split('.')[-1] in enabled_plugins:
                rv.append(plugin.run(**kwargs))
        return rv

    def walk_package(self, package):
        imported_package = __import__(package, fromlist=['blah'])

        for _, pluginname, ispkg in pkgutil.iter_modules(imported_package.__path__, imported_package.__name__ + '.'):
            if not ispkg:
                plugin_module = __import__(pluginname, fromlist=['blah'])
                clsmembers = inspect.getmembers(plugin_module, inspect.isclass)
                for (_, c) in clsmembers:
                    if issubclass(c, Plugin) & (c is not Plugin):
                        self.plugins.append(c(config=self.config[pluginname.split('.')[-1]] or {}))
            all_current_paths = []
            if isinstance(imported_package.__path__, str):
                all_current_paths.append(imported_package.__path__)
            else:
                all_current_paths.extend([x for x in imported_package.__path__])

            for pkg_path in all_current_paths:
                if pkg_path not in self.seen_paths:
                    self.seen_paths.append(pkg_path)

                    child_pkgs = [p for p in os.listdir(pkg_path) if os.path.isdir(os.path.join(pkg_path, p))]

                    for child_pkg in child_pkgs:
                        self.walk_package(package + '.' + child_pkg)



DB_FILE = '.redirects.yaml'


class RedirectFactory:
    def __init__(self, yamlloader=None, definitions_directory=None):
        self._definitions_directory = definitions_directory
        self._db_path = DB_FILE
        self._yamlloader = yamlloader
        self.db = self._yamlloader.loadyaml(self._db_path)
        if 'counter' not in self.db:
            self.db['counter'] = 38000
        if 'bindip' not in self.db:
            self.db['bindip'] = '44.128.191.10'
        if 'redirects' not in self.db:
            self.db['redirects'] = {}

    def allocate(self, via_service, target_service):
        if((via_service.tag, target_service.tag) not in self.db['redirects']):
            self.db['redirects'][(via_service.tag, target_service.tag)] = (self.db['bindip'], self.db['counter'])
            self.db['counter'] += 1
        return self.db['redirects'][(via_service.tag, target_service.tag)]

    def commit(self):
        self._yamlloader.dumpyaml(self.db, DB_FILE)




from setuptools import setup

name = 'emuserema'
version = '0.1.post1'
description = 'EMUSEREMA Multiprotocol Session and Redirect Manager'


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name=name,
    version=version,
    description=description,
    url='http://github.com/endreszabo/emuserema',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Systems Administration',
    ],
    author='Endre Szabo',
    author_email='emuserema@end.re',
    license='GPLv2',
    packages=['emuserema'],
    include_package_data=True,
    install_requires=[
        'ruamel.yaml',
    ],
    scripts=[
        'bin/viemuserema',
        'bin/emuserema'
    ],
    extras_require = {
        'ansible_inventory_plugin': ["ansible"]
    },
    zip_safe=False
)




from emuserema.services import SSHservice
from emuserema.plugin_manager import Plugin
from emuserema.utils import makedir_getfd


class WsshCommandsRenderer(Plugin):
    """This plugin is just the identity function: it returns the argument
    """
    def config(self):
        self.description = 'wssh launcher framework commands generator'

    def render_commands(self):
        with makedir_getfd(self._config['output_file']) as dmenu:
            print(self.render_kv(), file=dmenu)

    def render_kv(self):
        rv = []
        for key, service in self.services.items():
            #service=self.services[key]
            if isinstance(service, SSHservice):
                if ('_gui' not in service.kwargs or service.kwargs['_gui'] is True):
                    rv.append("%s%s" % (
                        '/'.join(service.path[1:]),
                        service.label)
                    )
        return("\n".join(rv))

    def run(self, **kwargs):
        """The actual implementation of the identity plugin is to just return the
        argument
        """
        self.services = kwargs['services']
        self.worlds = kwargs['worlds']
        self.render_commands()



from os import makedirs
from os.path import expanduser
from os import walk, path
from shutil import rmtree
from os import listdir, unlink


def cleanup_dir(target):
    for root, dirs, files in walk(target, topdown=False):
        for name in dirs:
            try:
                pass
                rmtree(path.join(root, name))
            except OSError as e:
                print("Ran into an exception during removal of %s: '%s'" % (path.join(root, name), e))


def cleanup_files(folder):
    for the_file in listdir(folder):
        file_path = path.join(folder, the_file)
        if not path.islink(file_path) and path.isfile(file_path):
            pass
            unlink(file_path)


def makedir_getfd(target):
    target = expanduser(target)
    makedirs(target[:target.rindex('/')], exist_ok=True)
    return open(target, 'w')


def service_paths_to_tree(services):
    #A simple recursion will do -- they said
    #https://codegolf.stackexchange.com/a/4485
    def build_nested_helper(path, service, container):
        segs = path
        head = segs[0]
        tail = segs[1:]
        if not tail:
            container.append({
                "text": "%s [%s]%s" % (service.tag, service.original_url,
                    service.via and " (via %s)" % service.via.tag or ''),
                "type": "website",
                "a_attr": {
                    "href": service.url
                }
            })
        else:
            #folder = {'service':head,'children': []}
            for item in container:
                if item['text'] == head:
                    build_nested_helper(tail, service, item['children'])
                    break
            else:
                folder = {
                    "type": "leaf",
                    "text": head,
                    "children": []
                }
                container.append(folder)
                build_nested_helper(tail, service, folder['children'])

    def build_nested(paths):
        container = []
        for item in paths:
            build_nested_helper(item[0], item[1], container)
        return container
    return build_nested(services)


def traverse(obj, item=None, callback=None):
    if item is None:
        item = []

    if isinstance(obj, dict):
        #if '_type' in obj:
        #    return obj['_type']
        value = {k: traverse(v, item + [k], callback)
                 for k, v in obj.items()}
    elif isinstance(obj, list):
        value = [traverse(elem, item + [[]], callback)
                 for elem in obj]
    else:
        value = obj

    if callback is None:
        return value
    else:
        return callback(item, value)


